#!/usr/bin/env python3

"""
Simple script to update the children properties information for
GESTrackElement-s that add children properties all the time
"""

import os
import textwrap
from itertools import chain

import gi

gi.require_version("Gst", "1.0")
gi.require_version("GObject", "2.0")
gi.require_version("GES", "1.0")

from gi.repository import Gst, GES, GObject  # noqa: E402

overrides = {
    "GstFramePositioner": False,
    "GstBaseTextOverlay": "GstBaseTextOverlay",  # Use the actual class name for proper linking
    "GstVideoDirection": "GstVideoDirection",     # Use the actual interface name for proper linking
    "GESVideoTestSource": "GESVideoTestSource",
    "GESVideoTransition": "GESVideoTransition",
}


def get_enum_values_from_property(prop):
    """
    Get enum values using proper introspection for newer pygobject versions.
    """
    enum_class = prop.enum_class

    # Get the enum type from the default value
    try:
        default_val = prop.get_default_value()
        enum_type = type(default_val)
    except Exception:
        # Fallback: create enum values as integers
        return [(i, f"Value_{i}", f"value-{i}") for i in range(enum_class.n_values)]

    values = []

    # In some cases there are 'gaps' like in the GstBaseTextOverlayHAlign where
    # value 3 is GST_BASE_TEXT_OVERLAY_HALIGN_UNUSED - so we need to compensate
    # for the gap
    n_values = enum_class.n_values
    for i in range(n_values):
        try:
            enum_type(i)
        except Exception:
            n_values += 1

    # Get all enum values by creating them with integer values
    for i in range(n_values):
        try:
            # For other values, create them using the enum type
            enum_val = enum_type(i)
            values.append((i, enum_val.value_name, enum_val.value_nick))
        except Exception:
            # Fallback for any failed value
            if i == 0:
                # Try to get first value info
                try:
                    first_val = enum_class.values
                    values.append((first_val.value, first_val.value_name, first_val.value_nick))
                except Exception:
                    continue
            else:
                continue

    return values


if __name__ == "__main__":
    Gst.init(None)
    GES.init()

    os.chdir(os.path.realpath(os.path.dirname(__file__)))
    tl = GES.Timeline.new_audio_video()
    layer = tl.append_layer()

    elements = []

    def add_clip(c, add=True, override_name=None):
        c.props.duration = Gst.SECOND
        c.props.start = layer.get_duration()
        layer.add_clip(c)
        if add:
            elements.extend(c.children)
        else:
            if override_name:
                elements.append((c, override_name))
            else:
                elements.append(c)

    add_clip(GES.UriClipAsset.request_sync(Gst.filename_to_uri(
        os.path.join("../../", "tests/check/assets/audio_video.ogg"))).extract())
    add_clip(GES.TestClip.new())
    add_clip(GES.TitleClip.new())

    add_clip(GES.SourceClip.new_time_overlay(), False, "GESTimeOverlaySourceClip")
    add_clip(GES.TransitionClip.new_for_nick("crossfade"), False)

    for element in elements:
        if isinstance(element, tuple):
            element, gtype = element
        else:
            gtype = element.__gtype__.name
        with open(gtype + '-children-props.md', 'w') as f:
            for prop in GES.TimelineElement.list_children_properties(element):
                prefix = '#### `%s`\n\n' % (prop.name)

                prefix_len = len(prefix)
                lines = [i for i in chain.from_iterable([textwrap.wrap(t, width=80) for t in prop.blurb.split('\n')])]

                doc = prefix + lines[0]

                if GObject.type_is_a(prop, GObject.ParamSpecEnum.__gtype__):
                    lines += ["", "Valid values:"]
                    try:
                        # Try the deprecated method first for backward compatibility
                        for value in prop.enum_class.__enum_values__.values():
                            lines.append("  - **%s** (%d) – %s" % (value.value_name,
                                                                   int(value), value.value_nick))
                    except AttributeError:
                        # Handle newer pygobject versions - use proper introspection
                        enum_values = get_enum_values_from_property(prop)
                        for value, name, nick in enum_values:
                            lines.append("  - **%s** (%d) – %s" % (name, value, nick))
                else:
                    lines += ["", "Value type: #" + prop.value_type.name]

                typename = overrides.get(prop.owner_type.name, None)
                if typename is not False:
                    if typename is None:
                        if GObject.type_is_a(prop.owner_type, Gst.Element):
                            typename = GObject.new(prop.owner_type).get_factory().get_name()
                    lines += ["", "See #%s:%s" % (typename, prop.name)]

                if len(lines) > 1:
                    doc += '\n'
                    doc += '\n'.join(lines[1:])

                print(doc + "\n", file=f)



import os
import gi.overrides

LOCAL_OVERRIDE_PATH = "gst-editing-services/bindings/python/gi/overrides/"
FILE = os.path.realpath(__file__)
if not gi.overrides.__path__[0].endswith(LOCAL_OVERRIDE_PATH):
    local_overrides = None
    # our overrides don't take precedence, let's fix it
    for i, path in enumerate(gi.overrides.__path__):
        if path.endswith(LOCAL_OVERRIDE_PATH):
            local_overrides = path

    if local_overrides:
        gi.overrides.__path__.remove(local_overrides)
    else:
        local_overrides = os.path.abspath(os.path.join(FILE, "../../../../../", LOCAL_OVERRIDE_PATH))

    gi.overrides.__path__.insert(0, local_overrides)

# Execute previously set sitecustomize.py script if it existed
if os.environ.get("GST_ENV"):
    old_sitecustomize = os.path.join(os.path.dirname(__file__),
                                    "old.sitecustomize.gstuninstalled.py")
    if os.path.exists(old_sitecustomize):
        exec(compile(open(old_sitecustomize).read(), old_sitecustomize, 'exec'))



#!/usr/bin/env python3
#
# Copyright (c) 2013, Thibault Saunier <thibault.saunier@collabora.com>
# Copyright (c) 2020, Igalia S.L
#    Author: Thibault Saunier <tsaunier@igalia.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
# Boston, MA 02110-1301, USA.

import os
import sys
import tempfile
import urllib.parse
import subprocess
from launcher import utils
from urllib.parse import unquote
import xml.etree.ElementTree as ET
from launcher.baseclasses import GstValidateTest, TestsManager, ScenarioManager, MediaFormatCombination, \
    MediaDescriptor, GstValidateEncodingTestInterface

GES_DURATION_TOLERANCE = utils.GST_SECOND / 2

GES_LAUNCH_COMMAND = "ges-launch-1.0"
if "win32" in sys.platform:
    GES_LAUNCH_COMMAND += ".exe"


GES_ENCODING_TARGET_COMBINATIONS = [
    MediaFormatCombination("ogg", "vorbis", "theora"),
    MediaFormatCombination("ogg", "opus", "theora"),
    MediaFormatCombination("webm", "vorbis", "vp8"),
    MediaFormatCombination("webm", "opus", "vp8"),
    MediaFormatCombination("mp4", "aac", "h264"),
    MediaFormatCombination("mp4", "ac3", "h264"),
    MediaFormatCombination("quicktime", "aac", "jpeg"),
    MediaFormatCombination("mkv", "opus", "h264"),
    MediaFormatCombination("mkv", "vorbis", "h264"),
    MediaFormatCombination("mkv", "opus", "jpeg"),
    MediaFormatCombination("mkv", "vorbis", "jpeg")
]


def quote_uri(uri):
    """
    Encode a URI/path according to RFC 2396, without touching the file:/// part.
    """
    # Split off the "file:///" part, if present.
    parts = urllib.parse.urlsplit(uri, allow_fragments=False)
    # Make absolutely sure the string is unquoted before quoting again!
    raw_path = unquote(parts.path)
    return utils.path2url(raw_path)


class XgesProjectDescriptor(MediaDescriptor):
    def __init__(self, uri):
        super(XgesProjectDescriptor, self).__init__()

        self._uri = uri
        self._xml_path = utils.url2path(uri)
        self._root = ET.parse(self._xml_path)
        self._duration = None

    def get_media_filepath(self):
        return self._xml_path

    def get_path(self):
        return self._xml_path

    def get_caps(self):
        raise NotImplemented

    def get_uri(self):
        return self._uri

    def get_duration(self):
        if self._duration:
            return self._duration

        for l in self._root.iter():
            if l.tag == "timeline":
                self._duration=int(l.attrib['metadatas'].split("duration=(guint64)")[1].split(" ")[0].split(";")[0])
                break

        if not self._duration:
            self.error("%s does not have duration! (setting 2mins)" % self._uri)
            self._duration = 2 * 60

        return self._duration

    def get_protocol(self):
        return Protocols.FILE

    def is_seekable(self):
        return True

    def is_image(self):
        return False

    def get_num_tracks(self, track_type):
        num_tracks = 0
        for l in self._root.iter():
            if l.tag == "track":
                if track_type in l.attrib["caps"]:
                    num_tracks += 1
        return num_tracks


class GESTest(GstValidateTest):
    def __init__(self, classname, options, reporter, project, scenario=None,
                 combination=None, expected_failures=None, nest=False, testfile=None):

        super(GESTest, self).__init__(GES_LAUNCH_COMMAND, classname, options, reporter,
                                      scenario=scenario)

        self.project = project
        self.nested = nest
        self.testfile = testfile

    def set_sample_paths(self):
        if self.testfile:
            # testfile should be self contained
            return

        if not self.options.paths:
            if self.options.disable_recurse:
                return
            if self.project:
                paths = [os.path.dirname(self.project.get_media_filepath())]
            else:
                paths = []
        else:
            paths = self.options.paths

        if not isinstance(paths, list):
            paths = [paths]

        for path in paths:
            # We always want paths separator to be cut with '/' for ges-launch
            path = path.replace("\\", "/")
            if not self.options.disable_recurse:
                self.add_arguments("--ges-sample-path-recurse", quote_uri(path))
                self.add_arguments("--ges-sample-path-recurse", quote_uri(self.options.projects_paths))
            else:
                self.add_arguments("--ges-sample-paths", quote_uri(path))

    def set_sink_args(self):
        if self.testfile:
            # testfile should be self contained and --mute should give required infos.
            if self.options.mute:
                self.add_arguments("--mute")
            return

        if self.options.mute:
            needs_clock = self.scenario.needs_clock_sync() \
                if self.scenario else False
            audiosink = utils.get_fakesink_for_media_type("audio", needs_clock)
            videosink = utils.get_fakesink_for_media_type("video", needs_clock)
        else:
            audiosink = 'autoaudiosink'
            videosink = 'autovideosink'
        self.add_arguments("--videosink", videosink + " name=videosink")
        self.add_arguments("--audiosink", audiosink + " name=audiosink")

    def build_arguments(self):
        GstValidateTest.build_arguments(self)

        self.set_sink_args()
        self.set_sample_paths()

        if self.project:
            assert self.testfile is None
            if self.nested:
                self.add_arguments("+clip", self.project.get_uri())
            else:
                self.add_arguments("-l", self.project.get_uri())
        elif self.testfile:
            self.add_arguments("--set-test-file", self.testfile)

class GESPlaybackTest(GESTest):
    def __init__(self, classname, options, reporter, project, scenario,nest):
        super(GESPlaybackTest, self).__init__(classname, options, reporter,
                                      project, scenario=scenario, nest=nest)

    def get_current_value(self):
        return self.get_current_position()

class GESScenarioTest(GESTest):
    def __init__(self, classname, options, reporter, scenario):
        super().__init__(classname, options, reporter, None, scenario=scenario)

    def build_arguments(self):
        super().build_arguments()
        self.add_arguments("--set-scenario", self.scenario.path)

    def get_subproc_env(self):
        scenario = self.scenario
        self.scenario = None
        res = super().get_subproc_env()
        self.scenario = scenario

        return res

    def get_current_value(self):
        return self.get_current_position()


class GESRenderTest(GESTest, GstValidateEncodingTestInterface):
    def __init__(self, classname, options, reporter, project, combination):
        GESTest.__init__(self, classname, options, reporter, project)

        GstValidateEncodingTestInterface.__init__(self, combination, self.project)

    def build_arguments(self):
        GESTest.build_arguments(self)
        self._set_rendering_info()

    def run_external_checks(self):
        reference_file_path = urllib.parse.urlsplit(self.media_descriptor.get_uri()).path + ".expected_result"
        if os.path.exists(reference_file_path):
            self.run_iqa_test(utils.path2url(reference_file_path))

    def _set_rendering_info(self):
        self.dest_file = path = os.path.join(self.options.dest,
                                             self.classname.replace(".render.", os.sep).
                                             replace(".", os.sep))
        utils.mkdir(os.path.dirname(urllib.parse.urlsplit(self.dest_file).path))
        if not utils.isuri(self.dest_file):
            self.dest_file = utils.path2url(self.dest_file)

        profile = self.get_profile()
        self.add_arguments("-f", profile, "-o", self.dest_file)

    def check_results(self):
        self.check_encoded_file()
        return GstValidateTest.check_results(self)

    def get_current_value(self):
        size = self.get_current_size()
        if size is None:
            return self.get_current_position()

        return size


class GESTestsManager(TestsManager):
    name = "ges"

    _scenarios = ScenarioManager()

    def __init__(self):
        super(GESTestsManager, self).__init__()

    def init(self):
        try:
            with tempfile.NamedTemporaryFile() as f:
                if "--set-scenario=" in subprocess.check_output([GES_LAUNCH_COMMAND, "--help"], stderr=f).decode():
                    return True
                else:
                    self.warning("Can not use ges-launch, it seems not to be compiled against"
                                " gst-validate")
        except subprocess.CalledProcessError as e:
            self.warning("Can not use ges-launch: %s" % e)
        except OSError as e:
            self.warning("Can not use ges-launch: %s" % e)

    def add_options(self, parser):
        group = parser.add_argument_group("GStreamer Editing Services specific option"
                            " and behaviours",
                            description="""
The GStreamer Editing Services launcher will be usable only if GES has been compiled against GstValidate
You can simply run scenarios specifying project as args. For example the following will run all available
and activated scenarios on project.xges:

    $gst-validate-launcher ges /some/ges/project.xges


Available options:""")
        group.add_argument("-P", "--projects-paths", dest="projects_paths",
                         default=os.path.join(utils.DEFAULT_GST_QA_ASSETS,
                                              "ges",
                                              "ges-projects"),
                         help="Paths in which to look for moved medias")
        group.add_argument("--ges-scenario-paths", dest="scenarios_path",
                         default=None,
                         help="Paths in which to look for moved medias")
        group.add_argument("-r", "--disable-recurse-paths", dest="disable_recurse",
                         default=False, action="store_true",
                         help="Whether to recurse into paths to find medias")

    def set_settings(self, options, args, reporter):
        TestsManager.set_settings(self, options, args, reporter)
        self._scenarios.config = self.options

        try:
            os.makedirs(utils.url2path(options.dest)[0])
        except OSError:
            pass

    def list_tests(self):
        return self.tests

    def register_defaults(self, project_paths=None, scenarios_path=None):
        projects = list()
        all_scenarios = {}
        if not self.args:
            if project_paths == None:
                path = self.options.projects_paths
            else:
                path = project_paths

            for root, dirs, files in os.walk(path):
                for f in files:
                    if not f.endswith(".xges"):
                        continue
                    projects.append(utils.path2url(os.path.join(path, root, f)))

            if self.options.scenarios_path:
                scenarios_path = self.options.scenarios_path

            if scenarios_path:
                for root, dirs, files in os.walk(scenarios_path):
                    for f in files:
                        name, ext = os.path.splitext(f)
                        f = os.path.join(root, f)
                        if ext == ".validatetest":
                            fpath = os.path.abspath(os.path.join(root, f))
                            pathname = os.path.abspath(os.path.join(root, name))
                            name = pathname.replace(os.path.commonpath([scenarios_path, root]), '').replace('/', '.')
                            self.add_test(GESTest('test' + name,
                                                  self.options,
                                                  self.reporter,
                                                  None,
                                                  testfile=fpath))
                            continue
                        elif ext != ".scenario":
                            continue
                        config = f + ".config"
                        if not os.path.exists(config):
                            config = None
                        all_scenarios[f] = config
        else:
            for proj_uri in self.args:
                if not utils.isuri(proj_uri):
                    proj_uri = utils.path2url(proj_uri)

                if os.path.exists(proj_uri):
                    projects.append(proj_uri)

        if self.options.long_limit != 0:
            scenarios = ["none",
                         "scrub_forward_seeking",
                         "scrub_backward_seeking"]
        else:
            scenarios = ["play_15s",
                         "scrub_forward_seeking_full",
                         "scrub_backward_seeking_full"]
        for proj_uri in projects:
            # First playback casses
            project = XgesProjectDescriptor(proj_uri)
            for scenario_name in scenarios:
                scenario = self._scenarios.get_scenario(scenario_name)
                if scenario is None:
                    continue

                if scenario.get_min_media_duration() >= (project.get_duration() / utils.GST_SECOND):
                    continue

                classname = "playback.%s.%s" % (scenario.name,
                                                    os.path.basename(proj_uri).replace(".xges", ""))
                self.add_test(GESPlaybackTest(classname,
                                              self.options,
                                              self.reporter,
                                              project,
                                              scenario=scenario,
                                              nest=False))
                #For nested timelines
                classname = "playback.nested.%s.%s" % (scenario.name,
                                                    os.path.basename(proj_uri).replace(".xges", ""))
                self.add_test(GESPlaybackTest(classname,
                                              self.options,
                                              self.reporter,
                                              project,
                                              scenario=scenario,
                                              nest=True))

            # And now rendering casses
            for comb in GES_ENCODING_TARGET_COMBINATIONS:
                classname = "render.%s.%s" % (str(comb).replace(' ', '_'),
                                                  os.path.splitext(os.path.basename(proj_uri))[0])
                self.add_test(GESRenderTest(classname, self.options,
                                            self.reporter, project,
                                            combination=comb)
                                  )
        if all_scenarios:
            for scenario in self._scenarios.discover_scenarios(list(all_scenarios.keys())):
                config = all_scenarios[scenario.path]
                classname = "scenario.%s" % scenario.name
                test = GESScenarioTest(classname, self.options, self.reporter, scenario=scenario)
                if config:
                    test.add_validate_config(config)
                self.add_test(test)


# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Thibault Saunier <tsaunier@igalia.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
# Boston, MA 02110-1301, USA.

from . import overrides_hack

import os
import gi
import tempfile

gi.require_version("Gst", "1.0")
gi.require_version("GES", "1.0")
gi.require_version("GstPbutils", "1.0")

from gi.repository import Gst  # noqa
from gi.repository import GstPbutils  # noqa
from gi.repository import GLib  # noqa
from gi.repository import GES  # noqa
import unittest  # noqa
from unittest import mock

from . import common
from .common import GESSimpleTimelineTest  # noqa

Gst.init(None)
GES.init()


class TestTimeline(GESSimpleTimelineTest):

    def test_request_relocated_assets_sync(self):
        path = os.path.join(__file__, "../../../", "png.png")
        with self.assertRaises(GLib.Error):
            GES.UriClipAsset.request_sync(Gst.filename_to_uri(path))

        GES.add_missing_uri_relocation_uri(Gst.filename_to_uri(os.path.join(__file__, "../../assets")), False)
        path = os.path.join(__file__, "../../", "png.png")
        self.assertEqual(GES.UriClipAsset.request_sync(Gst.filename_to_uri(path)).props.id,
            Gst.filename_to_uri(os.path.join(__file__, "../../assets/png.png")))

    def test_request_relocated_twice(self):
        GES.add_missing_uri_relocation_uri(Gst.filename_to_uri(os.path.join(__file__, "../../")), True)
        proj = GES.Project.new()

        asset = proj.create_asset_sync("file:///png.png", GES.UriClip)
        self.assertIsNotNone(asset)
        asset = proj.create_asset_sync("file:///png.png", GES.UriClip)
        self.assertIsNotNone(asset)

    @unittest.skipUnless(*common.can_generate_assets())
    def test_reload_asset(self):
        with common.created_video_asset(num_bufs=2, framerate="2/1") as uri:
            asset0 = GES.UriClipAsset.request_sync(uri)
            self.assertEqual(asset0.props.duration, Gst.SECOND)

            with common.created_video_asset(uri, 4, framerate="2/1") as uri:
                GES.Asset.needs_reload(GES.UriClip, uri)
                asset1 = GES.UriClipAsset.request_sync(uri)
                self.assertEqual(asset1.props.duration, 2 * Gst.SECOND)
                self.assertEqual(asset1, asset0)

                with common.created_video_asset(uri, 6, framerate="2/1") as uri:
                    mainloop = common.create_main_loop()

                    def asset_loaded_cb(_, res, mainloop):
                        asset2 = GES.Asset.request_finish(res)
                        self.assertEqual(asset2.props.duration, 3 * Gst.SECOND)
                        self.assertEqual(asset2, asset0)
                        mainloop.quit()

                    GES.Asset.needs_reload(GES.UriClip, uri)
                    GES.Asset.request_async(GES.UriClip, uri, None, asset_loaded_cb, mainloop)
                    mainloop.run()

    def test_asset_metadata_on_reload(self):
        mainloop = GLib.MainLoop()

        unused, xges_path = tempfile.mkstemp(suffix=".xges")
        project_uri = Gst.filename_to_uri(os.path.abspath(xges_path))

        asset_uri = Gst.filename_to_uri(os.path.join(__file__, "../../assets/audio_video.ogg"))
        xges = """<ges version='0.3'>
            <project properties='properties;' metadatas='metadatas;'>
                <ressources>
                    <asset id='%(uri)s' extractable-type-name='GESUriClip' properties='properties, supported-formats=(int)6, duration=(guint64)2003000000;' metadatas='metadatas, container-format=(string)Matroska, language-code=(string)und, application-name=(string)Lavc56.60.100, encoder-version=(uint)0, audio-codec=(string)Vorbis, nominal-bitrate=(uint)80000, bitrate=(uint)80000, video-codec=(string)&quot;On2 VP8&quot;, file-size=(guint64)223340, foo=(string)bar;' >
                    </asset>
                </ressources>
            </project>
            </ges>""" % {"uri": asset_uri}
        with open(xges_path, "w") as xges_file:
            xges_file.write(xges)

        def loaded_cb(project, timeline):
            asset = project.list_assets(GES.Extractable)[0]
            self.assertEqual(asset.get_meta("foo"), "bar")
            mainloop.quit()

        loaded_project = GES.Project(uri=project_uri, extractable_type=GES.Timeline)
        loaded_project.connect("loaded", loaded_cb)
        timeline = loaded_project.extract()
        mainloop.run()

    def test_asset_load_serialized_info(self):
        mainloop = GLib.MainLoop()

        serialized_infos = {}
        n_calls = 0
        n_cache_hits = 0

        def load_serialized_info_cb(_manager, uri):
            nonlocal n_calls, n_cache_hits, serialized_infos

            n_calls += 1
            res = serialized_infos.get(uri)
            if res:
                n_cache_hits += 1
            return res

        GES.DiscovererManager.get_default().connect("load-serialized-info",
                                                    load_serialized_info_cb)

        self.assertEqual(n_calls, 0)
        asset = GES.UriClipAsset.request_sync(Gst.filename_to_uri(os.path.join(__file__, "../../assets/audio_video.ogg")))
        self.assertEqual(n_calls, 1)
        self.assertEqual(n_cache_hits, 0)

        serialized_infos[asset.get_id()] = asset.get_info()

        # Clear the GES internal asset cache
        GES.deinit()
        GES.init()

        # Connect to the new manager, previous one was destroyed on deinit
        GES.DiscovererManager.get_default().connect("load-serialized-info",
                                                    load_serialized_info_cb)
        asset = GES.UriClipAsset.request_sync(Gst.filename_to_uri(os.path.join(__file__, "../../assets/audio_video.ogg")))
        self.assertEqual(n_calls, 2)
        self.assertEqual(n_cache_hits, 1)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Alexandru Băluț <alexandru.balut@gmail.com>
# Copyright (c) 2016, Thibault Saunier
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
# Boston, MA 02110-1301, USA.

from urllib.parse import urlparse
import gi

gi.require_version("Gst", "1.0")
gi.require_version("GES", "1.0")

from gi.repository import Gst  # noqa
from gi.repository import GES  # noqa
from gi.repository import GLib  # noqa
from gi.repository import GObject  # noqa
import contextlib  # noqa
import os  #noqa
import unittest  # noqa
import tempfile  # noqa

try:
    gi.require_version("GstTranscoder", "1.0")
    from gi.repository import GstTranscoder
except ValueError:
    GstTranscoder = None

Gst.init(None)
GES.init()


def create_main_loop():
    """Creates a MainLoop with a timeout."""
    mainloop = GLib.MainLoop()
    timed_out = False

    def timeout_cb(unused):
        nonlocal timed_out
        timed_out = True
        mainloop.quit()

    def run(timeout_seconds=5, until_empty=False):
        source = GLib.timeout_source_new_seconds(timeout_seconds)
        source.set_callback(timeout_cb)
        source.attach()
        if until_empty:
            GLib.idle_add(mainloop.quit)
        GLib.MainLoop.run(mainloop)
        source.destroy()
        if timed_out:
            raise Exception("Timed out after %s seconds" % timeout_seconds)

    mainloop.run = run
    return mainloop


def create_project(with_group=False, saved=False):
    """Creates a project with two clips in a group."""
    timeline = GES.Timeline.new_audio_video()
    layer = timeline.append_layer()

    if with_group:
        clip1 = GES.TitleClip()
        clip1.set_start(0)
        clip1.set_duration(10*Gst.SECOND)
        layer.add_clip(clip1)
        clip2 = GES.TitleClip()
        clip2.set_start(100 * Gst.SECOND)
        clip2.set_duration(10*Gst.SECOND)
        layer.add_clip(clip2)
        group = GES.Container.group([clip1, clip2])

    if saved:
        if isinstance(saved, str):
            suffix = "-%s.xges" % saved
        else:
            suffix = ".xges"
        uri = "file://%s" % tempfile.NamedTemporaryFile(suffix=suffix).name
        timeline.get_asset().save(timeline, uri, None, overwrite=True)

    return timeline


@contextlib.contextmanager
def created_project_file(xges):
    _, xges_path = tempfile.mkstemp(suffix=".xges")
    with open(xges_path, "w") as f:
        f.write(xges)

    yield Gst.filename_to_uri(os.path.abspath(xges_path))

    os.remove(xges_path)


def can_generate_assets():
    if GstTranscoder is None:
        return False, "GstTranscoder is not available"

    if not Gst.ElementFactory.make("testsrcbin"):
        return False, "testbinsrc is not available"

    return True, None


@contextlib.contextmanager
def created_video_asset(uri=None, num_bufs=30, framerate="30/1"):
    with tempfile.NamedTemporaryFile(suffix=".ogg") as f:
        if not uri:
            uri = Gst.filename_to_uri(f.name)
            name = f.name
        else:
            name = urlparse(uri).path
        pipe = Gst.parse_launch(f"videotestsrc num-buffers={num_bufs} ! video/x-raw,framerate={framerate} ! theoraenc ! oggmux ! filesink location={name}")
        pipe.set_state(Gst.State.PLAYING)
        assert pipe.get_bus().timed_pop_filtered(Gst.CLOCK_TIME_NONE, Gst.MessageType.EOS)
        pipe.set_state(Gst.State.NULL)

        yield uri


def get_asset_uri(name):
    python_tests_dir = os.path.dirname(os.path.abspath(__file__))
    assets_dir = os.path.join(python_tests_dir, "..", "assets")
    return Gst.filename_to_uri(os.path.join(assets_dir, name))


class GESTest(unittest.TestCase):

    def _log(self, func, format, *args):
        string = format
        if args:
            string = string % args[0]
        func(string)

    def log(self, format, *args):
        self._log(Gst.log, format, *args)

    def debug(self, format, *args):
        self._log(Gst.debug, format, *args)

    def info(self, format, *args):
        self._log(Gst.info, format, *args)

    def fixme(self, format, *args):
        self._log(Gst.fixme, format, *args)

    def warning(self, format, *args):
        self._log(Gst.warning, format, *args)

    def error(self, format, *args):
        self._log(Gst.error, format, *args)

    def check_clip_values(self, clip, start, in_point, duration):
        for elem in [clip] + clip.get_children(False):
            self.check_element_values(elem, start, in_point, duration)

    def check_element_values(self, element, start, in_point, duration):
        self.assertEqual(element.props.start, start, element)
        self.assertEqual(element.props.in_point, in_point, element)
        self.assertEqual(element.props.duration, duration, element)

    def assert_effects(self, clip, *effects):
        # Make sure there are no other effects.
        self.assertEqual(set(clip.get_top_effects()), set(effects))

        # Make sure their order is correct.
        indexes = [clip.get_top_effect_index(effect)
                   for effect in effects]
        self.assertEqual(indexes, list(range(len(effects))))

    def assertGESError(self, error, code, message=""):
        if error is None:
            raise AssertionError(
                "{}{}Received no error".format(message, message and ": "))
        if error.domain != "GES_ERROR":
            raise AssertionError(
                "{}{}Received error ({}) in domain {} rather than "
                "GES_ERROR".format(
                    message, message and ": ", error.message, error.domain))
        err_code = GES.Error(error.code)
        if err_code != code:
            raise AssertionError(
                "{}{}Received {} error ({}) rather than {}".format(
                    message, message and ": ", err_code.value_name,
                    error.message, code.value_name))

class GESSimpleTimelineTest(GESTest):

    def __init__(self, *args):
        self.track_types = [GES.TrackType.AUDIO, GES.TrackType.VIDEO]
        super(GESSimpleTimelineTest, self).__init__(*args)

    def timeline_as_str(self):
        res = "====== %s =======\n" % self.timeline
        for layer in self.timeline.get_layers():
            res += "Layer %04d: " % layer.get_priority()
            for clip in layer.get_clips():
                res += "{ %s }" % clip
            res += '\n------------------------\n'

        for group in self.timeline.get_groups():
            res += "GROUP %s :" % group
            for clip in group.get_children(False):
                res += " { %s }" % clip.props.name
            res += '\n'
        res += "================================\n"
        return res

    def print_timeline(self):
        print(self.timeline_as_str())

    def setUp(self):
        self.timeline = GES.Timeline.new()
        for track_type in self.track_types:
            self.assertIn(
                track_type, [GES.TrackType.AUDIO, GES.TrackType.VIDEO])
            if track_type == GES.TrackType.AUDIO:
                self.assertTrue(self.timeline.add_track(GES.AudioTrack.new()))
            else:
                self.assertTrue(self.timeline.add_track(GES.VideoTrack.new()))

        self.assertEqual(len(self.timeline.get_tracks()),
                         len(self.track_types))
        self.layer = self.timeline.append_layer()

    def add_clip(self, start, in_point, duration, asset_type=GES.TestClip):
        clip = GES.Asset.request(asset_type, None).extract()
        clip.props.start = start
        clip.props.in_point = in_point
        clip.props.duration = duration
        self.assertTrue(self.layer.add_clip(clip))

        return clip

    def append_clip(self, layer=0, asset_type=GES.TestClip, asset_id=None):
        while len(self.timeline.get_layers()) < layer + 1:
            self.timeline.append_layer()
        layer = self.timeline.get_layers()[layer]
        if asset_type == GES.UriClip:
            asset = GES.UriClipAsset.request_sync(asset_id)
        else:
            asset = GES.Asset.request(asset_type, asset_id)
        clip = asset.extract()
        clip.props.start = layer.get_duration()
        clip.props.duration = 10
        self.assertTrue(layer.add_clip(clip))

        return clip

    def assertElementAreEqual(self, ref, element):
        self.assertTrue(isinstance(element, type(ref)), "%s and %s do not have the same type!" % (ref, element))

        props = [p for p in ref.list_properties() if p.name not in ['name']
            and not GObject.type_is_a(p.value_type, GObject.Object)]
        for p in props:
            pname = p.name
            refval = GObject.Value()
            refval.init(p.value_type)
            refval.set_value(ref.get_property(pname))

            value = GObject.Value()
            value.init(p.value_type)
            value.set_value(element.get_property(pname))

            self.assertTrue(Gst.value_compare(refval, value) == Gst.VALUE_EQUAL,
                "%s are not equal: %s != %s\n    %s != %s" % (pname, value, refval, element, ref))

        if isinstance(ref, GES.TrackElement):
            self.assertElementAreEqual(ref.get_nleobject(), element.get_nleobject())
            return

        if not isinstance(ref, GES.Clip):
            return

        ttypes = [track.type for track in self.timeline.get_tracks()]
        for ttype in ttypes:
            if ttypes.count(ttype) > 1:
                self.warning("Can't deeply check %s and %s "
                    "(only one track per type supported %s %s found)" % (ref,
                    element, ttypes.count(ttype), ttype))
                return

        children = element.get_children(False)
        for ref_child in ref.get_children(False):
            ref_track = ref_child.get_track()
            if not ref_track:
                self.warning("Can't check %s as not in a track" % (ref_child))
                continue

            child = None
            for tmpchild in children:
                if not isinstance(tmpchild, type(ref_child)):
                    continue

                if ref_track.type != tmpchild.get_track().type:
                    continue

                if not isinstance(ref_child, GES.Effect):
                    child = tmpchild
                    break
                elif ref_child.props.bin_description == tmpchild.props.bin_description:
                    child = tmpchild
                    break

            self.assertIsNotNone(child, "Could not find equivalent child %s in %s(%s)" % (ref_child,
                element, children))

            self.assertElementAreEqual(ref_child, child)

    def check_reload_timeline(self):
        tmpf = tempfile.NamedTemporaryFile(suffix='.xges')
        uri = Gst.filename_to_uri(tmpf.name)
        self.assertTrue(self.timeline.save_to_uri(uri, None, True))
        project = GES.Project.new(uri)
        mainloop = create_main_loop()
        def loaded_cb(unused_project, unused_timeline):
            mainloop.quit()

        project.connect("loaded", loaded_cb)
        reloaded_timeline = project.extract()

        mainloop.run()
        self.assertIsNotNone(reloaded_timeline)

        layers = self.timeline.get_layers()
        reloaded_layers = reloaded_timeline.get_layers()
        self.assertEqual(len(layers), len(reloaded_layers))
        for layer, reloaded_layer in zip(layers, reloaded_layers):
            clips = layer.get_clips()
            reloaded_clips = reloaded_layer.get_clips()
            self.assertEqual(len(clips), len(reloaded_clips))
            for clip, reloaded_clip in zip(clips, reloaded_clips):
                self.assertElementAreEqual(clip, reloaded_clip)

        return reloaded_timeline

    def assertTimelineTopology(self, topology, groups=[]):
        res = []
        for layer in self.timeline.get_layers():
            layer_timings = []
            for clip in layer.get_clips():
                layer_timings.append(
                    (type(clip), clip.props.start, clip.props.duration))
                for child in clip.get_children(True):
                    self.assertEqual(child.props.start, clip.props.start)
                    self.assertEqual(child.props.duration, clip.props.duration)

            res.append(layer_timings)
        if topology != res:
            Gst.error(self.timeline_as_str())
            self.assertEqual(topology, res)

        timeline_groups = self.timeline.get_groups()
        if groups and timeline_groups:
            for i, group in enumerate(groups):
                self.assertEqual(set(group), set(timeline_groups[i].get_children(False)))
            self.assertEqual(len(timeline_groups), i + 1)

        return res


class GESTimelineConfigTest(GESTest):
    """
    Tests where all the configuration changes, snapping positions and
    auto-transitions are accounted for.
    """

    def setUp(self):
        timeline = GES.Timeline.new()
        self.timeline = timeline
        timeline.set_auto_transition(True)

        self.snap_occured = False
        self.snap = None

        def snap_started(tl, el1, el2, pos):
            if self.snap_occured:
                raise AssertionError(
                    "Previous snap {} not accounted for".format(self.snap))
            self.snap_occured = True
            if self.snap is not None:
                raise AssertionError(
                    "Previous snap {} not ended".format(self.snap))
            self.snap = (el1.get_parent(), el2.get_parent(), pos)

        def snap_ended(tl, el1, el2, pos):
            self.assertEqual(
                self.snap, (el1.get_parent(), el2.get_parent(), pos))
            self.snap = None

        timeline.connect("snapping-started", snap_started)
        timeline.connect("snapping-ended", snap_ended)

        self.lost_clips = []

        def unrecord_lost_clip(layer, clip):
            if clip in self.lost_clips:
                self.lost_clips.remove(clip)

        def record_lost_clip(layer, clip):
            self.lost_clips.append(clip)

        def layer_added(tl, layer):
            layer.connect("clip-added", unrecord_lost_clip)
            layer.connect("clip-removed", record_lost_clip)

        timeline.connect("layer-added", layer_added)

        self.clips = []
        self.auto_transitions = {}
        self.config = {}

    @staticmethod
    def new_config(start, duration, inpoint, maxduration, layer):
        return {"start": start, "duration": duration, "in-point": inpoint,
                "max-duration": maxduration, "layer": layer}

    def add_clip(self, name, layer, tracks, start, duration, inpoint=0,
                 maxduration=Gst.CLOCK_TIME_NONE, clip_type=GES.TestClip,
                 asset_id=None, effects=None):
        """
        Create a clip with the given @name and properties and add it to the
        layer of priority @layer to the tracks in @tracks. Also registers
        its expected configuration.
        """
        if effects is None:
            effects = []

        lay = self.timeline.get_layer(layer)
        while lay is None:
            self.timeline.append_layer()
            lay = self.timeline.get_layer(layer)

        asset = GES.Asset.request(clip_type, asset_id)
        clip = asset.extract()
        self.assertTrue(clip.set_name(name))
        # FIXME: would be better to use select-tracks-for-object
        # hack around the fact that we cannot use select-tracks-for-object
        # in python by setting start to large number to ensure no conflict
        # when adding a clip
        self.assertTrue(clip.set_start(10000))
        self.assertTrue(clip.set_duration(duration))
        self.assertTrue(clip.set_inpoint(inpoint))

        for effect in effects:
            self.assertTrue(clip.add(effect))

        if lay.add_clip(clip) != True:
            raise AssertionError(
                "Failed to add clip {} to layer {}".format(name, layer))

        # then remove the children not in the selected tracks, which may
        # now allow some clips to fully/triple overlap because they do
        # not share a track
        for child in clip.get_children(False):
            if child.get_track() not in tracks:
                clip.remove(child)

        # then move to the desired start
        prev_snap = self.timeline.get_snapping_distance()
        self.timeline.set_snapping_distance(0)
        self.assertTrue(clip.set_start(start))
        self.timeline.set_snapping_distance(prev_snap)

        self.assertTrue(clip.set_max_duration(maxduration))

        self.config[clip] = self.new_config(
            start, duration, inpoint, maxduration, layer)
        self.clips.append(clip)

        return clip

    def add_group(self, name, to_group):
        """
        Create a group with the given @name and the elements in @to_group.
        Also registers its expected configuration.
        """
        group = GES.Group.new()
        self.assertTrue(group.set_name(name))
        start = None
        end = None
        layer = None
        for element in to_group:
            if start is None:
                start = element.start
                end = element.start + element.duration
                layer = element.get_layer_priority()
            else:
                start = min(start, element.start)
                end = max(end, element.start + element.duration)
                layer = min(layer, element.get_layer_priority())
            self.assertTrue(group.add(element))

        self.config[group] = self.new_config(
            start, end - start, 0, Gst.CLOCK_TIME_NONE, layer)
        return group

    def register_auto_transition(self, clip1, clip2, track):
        """
        Register that we expect an auto-transition to exist between
        @clip1 and @clip2 in @track.
        """
        transition = self._find_transition(clip1, clip2, track)
        if transition is None:
            raise AssertionError(
                "{} and {} have no auto-transition in track {}".format(
                    clip1, clip2, track))
        if transition in self.auto_transitions.values():
            raise AssertionError(
                "Auto-transition between {} and {} in track {} already "
                "registered".format(clip1, clip2, track))
        key = (clip1, clip2, track)
        if key in self.auto_transitions:
            raise AssertionError(
                "Auto-transition already registered for {}".format(key))

        self.auto_transitions[key] = transition

    def add_video_track(self):
        track = GES.VideoTrack.new()
        self.assertTrue(self.timeline.add_track(track))
        return track

    def add_audio_track(self):
        track = GES.AudioTrack.new()
        self.assertTrue(self.timeline.add_track(track))
        return track

    def assertElementConfig(self, element, config):
        for prop in config:
            if prop == "layer":
                val = element.get_layer_priority()
            else:
                val = element.get_property(prop)

            if val != config[prop]:
                raise AssertionError("{} property {}: {} != {}".format(
                    element, prop, val, config[prop]))

    @staticmethod
    def _source_in_track(clip, track):
        if clip.find_track_element(track, GES.Source):
            return True
        return False

    def _find_transition(self, clip1, clip2, track):
        """find transition from earlier clip1 to later clip2"""
        if not self._source_in_track(clip1, track) or \
                not self._source_in_track(clip2, track):
            return None

        layer_prio = clip1.get_layer_priority()
        if layer_prio != clip2.get_layer_priority():
            return None

        if clip1.start >= clip2.start:
            return None

        start = clip2.start
        end = clip1.start + clip1.duration
        if start >= end:
            return None
        duration = end - start

        layer = self.timeline.get_layer(layer_prio)
        self.assertIsNotNone(layer)

        for clip in layer.get_clips():
            children = clip.get_children(False)
            if len(children) == 1:
                child = children[0]
            else:
                continue
            if isinstance(clip, GES.TransitionClip) and clip.start == start \
                    and clip.duration == duration and child.get_track() == track:
                return clip

        raise AssertionError(
            "No auto-transition between {} and {} in track {}".format(
                clip1, clip2, track))

    def _transition_between(self, new, existing, clip1, clip2, track):
        if clip1.start < clip2.start:
            entry = (clip1, clip2, track)
        else:
            entry = (clip2, clip1, track)
        trans = self._find_transition(*entry)

        if trans is None:
            return

        if entry in new:
            new.remove(entry)
            self.auto_transitions[entry] = trans
        elif entry in existing:
            existing.remove(entry)
            expect = self.auto_transitions[entry]
            if trans != expect:
                raise AssertionError(
                    "Auto-transition between {} and {} in track {} changed "
                    "from {} to {}".format(
                        clip1, clip2, track, expect, trans))
        else:
            raise AssertionError(
                "Unexpected transition found between {} and {} in track {}"
                "".format(clip1, clip2, track))

    def assertTimelineConfig(
            self, new_props=None, snap_position=None, snap_froms=None,
            snap_tos=None, new_transitions=None, lost_transitions=None):
        """
        Check that the timeline configuration has only changed by the
        differences present in @new_props.
        Check that a snap occurred at @snap_position between one of the
        clips in @snap_froms and one of the clips in @snap_tos.
        Check that all new transitions in the timeline are present in
        @new_transitions.
        Checl that all the transitions that were lost are in
        @lost_transitions.
        """
        if new_props is None:
            new_props = {}
        if snap_froms is None:
            snap_froms = []
        if snap_tos is None:
            snap_tos = []
        if new_transitions is None:
            new_transitions = []
        if lost_transitions is None:
            lost_transitions = []

        for element, config in new_props.items():
            if element not in self.config:
                self.config[element] = {}

            for prop in config:
                self.config[element][prop] = new_props[element][prop]

        for element, config in self.config.items():
            self.assertElementConfig(element, config)

        # check that snapping occurred
        snaps = []
        for snap_from in snap_froms:
            for snap_to in snap_tos:
                snaps.append((snap_from, snap_to, snap_position))

        if self.snap is None:
            if snaps:
                raise AssertionError(
                    "No snap occurred, but expected a snap in {}".format(snaps))
        elif not snaps:
            if self.snap_occured:
                raise AssertionError(
                    "Snap {} occurred, but expected no snap".format(self.snap))
        elif self.snap not in snaps:
            raise AssertionError(
                "Snap {} occurred, but expected a snap in {}".format(
                    self.snap, snaps))
        self.snap_occured = False

        # check that lost transitions are not part of the layer
        for clip1, clip2, track in lost_transitions:
            key = (clip1, clip2, track)
            if key not in self.auto_transitions:
                raise AssertionError(
                    "No such auto-transition between {} and {} in track {} "
                    "is registered".format(clip1, clip2, track))
            # make sure original transition was removed from the layer
            trans = self.auto_transitions[key]
            if trans not in self.lost_clips:
                raise AssertionError(
                    "The auto-transition {} between {} and {} track {} was "
                    "not removed from the layers, but expect it to be lost"
                    "".format(trans, clip1, clip2, track))
            self.lost_clips.remove(trans)
            # make sure a new one wasn't created
            trans = self._find_transition(clip1, clip2, track)
            if trans is not None:
                raise AssertionError(
                    "Found auto-transition between {} and {} in track {} "
                    "is present, but expected it to be lost".format(
                        clip1, clip2, track))
            # since it was lost, remove it
            del self.auto_transitions[key]

        # check that all lost clips are accounted for
        if self.lost_clips:
            raise AssertionError(
                "Clips were lost that are not accounted for: {}".format(
                    self.lost_clips))

        # check that all other transitions are either existing ones or
        # new ones
        new = set(new_transitions)
        existing = set(self.auto_transitions.keys())
        for i, clip1 in enumerate(self.clips):
            for clip2 in self.clips[i+1:]:
                for track in self.timeline.get_tracks():
                    self._transition_between(
                        new, existing, clip1, clip2, track)

        # make sure we are not missing any expected transitions
        if new:
            raise AssertionError(
                "Did not find new transitions for {}".format(new))
        if existing:
            raise AssertionError(
                "Did not find existing transitions for {}".format(existing))

        # make sure there aren't any clips we are unaware of
        transitions = self.auto_transitions.values()
        for layer in self.timeline.get_layers():
            for clip in layer.get_clips():
                if clip not in self.clips and clip not in transitions:
                    raise AssertionError("Unknown clip {}".format(clip))

    def assertEdit(self, element, layer, mode, edge, position, snap,
                   snap_froms, snap_tos, new_props, new_transitions,
                   lost_transitions):
        if not element.edit_full(layer, mode, edge, position):
            raise AssertionError(
                "Edit of {} to layer {}, mode {}, edge {}, at position {} "
                "failed when a success was expected".format(
                    element, layer, mode, edge, position))
        self.assertTimelineConfig(
            new_props=new_props, snap_position=snap, snap_froms=snap_froms,
            snap_tos=snap_tos, new_transitions=new_transitions,
            lost_transitions=lost_transitions)

    def assertFailEdit(self, element, layer, mode, edge, position, err_code):
        res = None
        error = None
        try:
            res = element.edit_full(layer, mode, edge, position)
        except GLib.Error as exception:
            error = exception

        if err_code is None:
            if res is not False:
                raise AssertionError(
                    "Edit of {} to layer {}, mode {}, edge {}, at "
                    "position {} succeeded when a failure was expected"
                    "".format(
                        element, layer, mode, edge, position))
            if error is not None:
                raise AssertionError(
                    "Edit of {} to layer {}, mode {}, edge {}, at "
                    "position {} did produced an error when none was "
                    "expected".format(
                        element, layer, mode, edge, position))
        else:
            self.assertGESError(
                error, err_code,
                "Edit of {} to layer {}, mode {}, edge {}, at "
                "position {}".format(element, layer, mode, edge, position))
        # should be no change or snapping if edit fails
        self.assertTimelineConfig()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2015, Thibault Saunier
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
# Boston, MA 02110-1301, USA.

from . import overrides_hack

import gi

gi.require_version("Gst", "1.0")
gi.require_version("GES", "1.0")

from gi.repository import Gst  # noqa
from gi.repository import GES  # noqa

from . import common  # noqa

import unittest  # noqa
from unittest import mock

Gst.init(None)
GES.init()


class TestGroup(common.GESSimpleTimelineTest):

    def testCopyGroup(self):
        clip1 = GES.TestClip.new()
        clip1.props.duration = 10

        self.layer.add_clip(clip1)

        self.assertEqual(len(clip1.get_children(False)), 2)

        group = GES.Group.new()
        self.assertTrue(group.add(clip1))

        self.assertEqual(len(group.get_children(False)), 1)

        group_copy = group.copy(True)
        self.assertEqual(len(group_copy.get_children(False)), 0)

        self.assertTrue(group_copy.paste(10))
        clips = self.layer.get_clips()
        self.assertEqual(len(clips), 2)
        self.assertEqual(clips[1].props.start, 10)

        clips[1].edit([], 1, GES.EditMode.EDIT_NORMAL, GES.Edge.EDGE_NONE, 10)
        clips = self.layer.get_clips()
        self.assertEqual(len(clips), 1)

    def testPasteChangedGroup(self):
        clip1 = GES.TestClip.new()
        clip1.props.duration = 10

        clip2 = GES.TestClip.new()
        clip2.props.start = 20
        clip2.props.duration = 10

        self.layer.add_clip(clip1)
        self.layer.add_clip(clip2)

        self.assertEqual(len(clip1.get_children(False)), 2)

        group = GES.Group.new()
        self.assertTrue(group.add(clip1))

        self.assertEqual(len(group.get_children(False)), 1)

        group_copy = group.copy(True)
        self.assertEqual(len(group_copy.get_children(False)), 0)

        self.assertTrue(group.add(clip2))
        self.assertEqual(len(group.get_children(False)), 2)
        self.assertEqual(len(group_copy.get_children(False)), 0)

        self.assertTrue(group_copy.paste(10))
        clips = self.layer.get_clips()
        self.assertEqual(len(clips), 3)
        self.assertEqual(clips[1].props.start, 10)

    def testPasteChangedGroup(self):
        clip1 = GES.TestClip.new()
        clip1.props.duration = 10

        clip2 = GES.TestClip.new()
        clip2.props.start = 20
        clip2.props.duration = 10

        self.layer.add_clip(clip1)
        self.layer.add_clip(clip2)

        self.assertEqual(len(clip1.get_children(False)), 2)

        group = GES.Group.new()
        self.assertTrue(group.add(clip1))

        self.assertEqual(len(group.get_children(False)), 1)

        group_copy = group.copy(True)
        self.assertEqual(len(group_copy.get_children(False)), 0)

        self.assertTrue(group.add(clip2))
        self.assertEqual(len(group.get_children(False)), 2)
        self.assertEqual(len(group_copy.get_children(False)), 0)

        self.assertTrue(group_copy.paste(10))
        clips = self.layer.get_clips()
        self.assertEqual(len(clips), 3)
        self.assertEqual(clips[1].props.start, 10)

    def test_move_clips_between_layers_with_auto_transition(self):
        self.timeline.props.auto_transition = True
        layer2 = self.timeline.append_layer()
        clip1 = GES.TestClip.new()
        clip1.props.start = 0
        clip1.props.duration = 30

        clip2 = GES.TestClip.new()
        clip2.props.start = 20
        clip2.props.duration = 20

        self.layer.add_clip(clip1)
        self.layer.add_clip(clip2)

        clips = self.layer.get_clips()
        self.assertEqual(len(clips), 4)
        self.assertEqual(layer2.get_clips(), [])

        group = GES.Container.group(clips)
        self.assertIsNotNone(group)

        self.assertTrue(clip1.edit(
            self.timeline.get_layers(), 1, GES.EditMode.EDIT_NORMAL, GES.Edge.EDGE_NONE, 0))
        self.assertEqual(self.layer.get_clips(), [])

        clips = layer2.get_clips()
        self.assertEqual(len(clips), 4)

    def test_remove_emits_signal(self):
        clip1 = GES.TestClip.new()
        self.layer.add_clip(clip1)

        group = GES.Group.new()
        child_removed_cb = mock.Mock()
        group.connect("child-removed", child_removed_cb)

        group.add(clip1)
        group.remove(clip1)
        child_removed_cb.assert_called_once_with(group, clip1)

        group.add(clip1)
        child_removed_cb.reset_mock()
        group.ungroup(recursive=False)
        child_removed_cb.assert_called_once_with(group, clip1)

    def test_loaded_project_has_groups(self):
        mainloop = common.create_main_loop()
        timeline = common.create_project(with_group=True, saved=True)
        layer, = timeline.get_layers()
        group, = timeline.get_groups()
        self.assertEqual(len(layer.get_clips()), 2)
        for clip in layer.get_clips():
            self.assertEqual(clip.get_parent(), group)

        # Reload the project, check the group.
        project = GES.Project.new(uri=timeline.get_asset().props.uri)

        loaded_called = False
        def loaded(unused_project, unused_timeline):
            nonlocal loaded_called
            loaded_called = True
            mainloop.quit()
        project.connect("loaded", loaded)

        timeline = project.extract()

        mainloop.run()
        self.assertTrue(loaded_called)

        layer, = timeline.get_layers()
        group, = timeline.get_groups()
        self.assertEqual(len(layer.get_clips()), 2)
        for clip in layer.get_clips():
            self.assertEqual(clip.get_parent(), group)

    def test_moving_group_with_transition(self):
        self.timeline.props.auto_transition = True
        clip1 = GES.TestClip.new()
        clip1.props.start = 0
        clip1.props.duration = 30

        clip2 = GES.TestClip.new()
        clip2.props.start = 20
        clip2.props.duration = 20

        self.layer.add_clip(clip1)
        self.layer.add_clip(clip2)

        clips = self.layer.get_clips()
        self.assertEqual(len(clips), 4)

        video_transition = None
        audio_transition = None
        for clip in clips:
            if isinstance(clip, GES.TransitionClip):
                if isinstance(clip.get_children(False)[0], GES.VideoTransition):
                    video_transition = clip
                else:
                    audio_transition = clip
        self.assertIsNotNone(audio_transition)
        self.assertIsNotNone(video_transition)

        self.assertEqual(video_transition.props.start, 20)
        self.assertEqual(video_transition.props.duration, 10)
        self.assertEqual(audio_transition.props.start, 20)
        self.assertEqual(audio_transition.props.duration, 10)

        group = GES.Container.group(clips)
        self.assertIsNotNone(group)

        self.assertTrue(clip2.edit(
            self.timeline.get_layers(), 0,
            GES.EditMode.EDIT_NORMAL, GES.Edge.EDGE_NONE, 25))
        clip2.props.start = 25

        clips = self.layer.get_clips()
        self.assertEqual(len(clips), 4)
        self.assertEqual(clip1.props.start, 5)
        self.assertEqual(clip1.props.duration, 30)
        self.assertEqual(clip2.props.start, 25)
        self.assertEqual(clip2.props.duration, 20)

        self.assertEqual(video_transition.props.start, 25)
        self.assertEqual(video_transition.props.duration, 10)
        self.assertEqual(audio_transition.props.start, 25)
        self.assertEqual(audio_transition.props.duration, 10)

    def test_moving_group_snapping_from_the_middle(self):
        self.track_types = [GES.TrackType.AUDIO]
        super().setUp()
        snapped_positions = []
        def snapping_started_cb(timeline, first_element, second_element,
                                position, snapped_positions):
            snapped_positions.append(position)

        self.timeline.props.snapping_distance = 5
        self.timeline.connect("snapping-started", snapping_started_cb,
                              snapped_positions)

        for start in range(0, 20, 5):
            clip = GES.TestClip.new()
            clip.props.start = start
            clip.props.duration = 5
            self.layer.add_clip(clip)

        clips = self.layer.get_clips()
        self.assertEqual(len(clips), 4)

        group = GES.Container.group(clips[1:3])
        self.assertIsNotNone(group)

        self.assertTimelineTopology([
            [
                (GES.TestClip, 0, 5),
                (GES.TestClip, 5, 5),
                (GES.TestClip, 10, 5),
                (GES.TestClip, 15, 5),
            ],
        ], groups=[clips[1:3]])

        self.assertEqual(clips[1].props.start, 5)
        self.assertEqual(clips[2].props.start, 10)
        clips[2].edit([], 0, GES.EditMode.EDIT_NORMAL, GES.Edge.EDGE_NONE, 11)

        self.assertEqual(snapped_positions[0], 5)
        self.assertTimelineTopology([
            [
                (GES.TestClip, 0, 5),
                (GES.TestClip, 5, 5),
                (GES.TestClip, 10, 5),
                (GES.TestClip, 15, 5),
            ],
        ], groups=[clips[1:3]])

    def test_rippling_with_group(self):
        self.track_types = [GES.TrackType.AUDIO]
        super().setUp()
        for _ in range(4):
            self.append_clip()

        snapped_positions = []
        def snapping_started_cb(timeline, first_element, second_element,
                                position, snapped_positions):
            snapped_positions.append(position)

        self.timeline.props.snapping_distance = 5
        self.timeline.connect("snapping-started", snapping_started_cb,
                              snapped_positions)

        clips = self.layer.get_clips()
        self.assertEqual(len(clips), 4)

        group_clips = clips[1:3]
        GES.Container.group(group_clips)
        self.assertTimelineTopology([
            [
                (GES.TestClip, 0, 10),
                (GES.TestClip, 10, 10),
                (GES.TestClip, 20, 10),
                (GES.TestClip, 30, 10),
            ],
        ], groups=[group_clips])

        self.assertFalse(clips[2].edit([], 0, GES.EditMode.EDIT_RIPPLE, GES.Edge.EDGE_NONE, 5))

        self.assertTimelineTopology([
            [
                (GES.TestClip, 0, 10),
                (GES.TestClip, 10, 10),
                (GES.TestClip, 20, 10),
                (GES.TestClip, 30, 10),
            ],
        ], groups=[group_clips])

        # Negative start...
        self.assertFalse(clips[2].edit([], 1, GES.EditMode.EDIT_RIPPLE, GES.Edge.EDGE_NONE, 1))
        self.assertTimelineTopology([
            [
                (GES.TestClip, 0, 10),
                (GES.TestClip, 10, 10),
                (GES.TestClip, 20, 10),
                (GES.TestClip, 30, 10),
            ],
        ], groups=[group_clips])

        self.assertTrue(clips[2].edit([], 1, GES.EditMode.EDIT_RIPPLE, GES.Edge.EDGE_NONE, 20))
        self.assertTimelineTopology([
            [
                (GES.TestClip, 0, 10),
            ],
            [
                (GES.TestClip, 10, 10),
                (GES.TestClip, 20, 10),
                (GES.TestClip, 30, 10),
            ],
        ], groups=[group_clips])

    def test_group_priority(self):
        self.track_types = [GES.TrackType.AUDIO]
        self.setUp()

        clip0 = self.append_clip()
        clip1 = self.append_clip(1)
        clip1.props.start = 20

        group = GES.Group.new()
        group.add(clip0)
        group.add(clip1)
        self.assertEqual(group.get_layer_priority(), 0)
        self.assertTimelineTopology([
            [
                (GES.TestClip, 0, 10),
            ],
            [
                (GES.TestClip, 20, 10),
            ]
        ], groups=[(clip0, clip1)])
        group.remove(clip0)
        self.assertEqual(group.get_layer_priority(), 1)

        clip1.edit(self.timeline.get_layers(), 2, GES.EditMode.EDIT_NORMAL, GES.Edge.EDGE_NONE, clip1.start)
        self.assertTimelineTopology([
            [
                (GES.TestClip, 0, 10),
            ],
            [ ],
            [
                (GES.TestClip, 20, 10),
            ]
        ], groups=[(clip1,)])

        self.assertEqual(group.get_layer_priority(), 2)
        self.assertTrue(clip1.edit(self.timeline.get_layers(), 0, GES.EditMode.EDIT_NORMAL, GES.Edge.EDGE_NONE, clip1.start))

        self.assertTimelineTopology([
            [
                (GES.TestClip, 0, 10),
                (GES.TestClip, 20, 10),
            ],
            [ ],
            [ ]
        ], groups=[(clip1,)])
